"""config module"""

import configparser
import logging

logger = logging.getLogger(__name__)


class Configuration:
    """Class containing all the configuration parameters of the program"""

    def __init__(self, config_file_path: str):
        self._config: dict = {}
        self._config_parser: configparser = configparser.ConfigParser()

        try:
            with open(config_file_path, "r", encoding="UTF-8") as config_file:
                self._config_parser.read_file(config_file)
        except (OSError, IOError) as e:
            err_msg = f"Error reading configuration from file {config_file_path}"
            raise ValueError(err_msg, e) from e

        try:
            logger.info("Reading configuration...")
            self.__read_language_configuration()
            self.__read_twitter_configuration()
            self.__read_postgresql_configuration()
            logger.info("Configuration correctly read.")
        except Exception as e:
            err_msg = (
                f"Error reading configuration parameters from file {config_file_path}"
            )
            raise ValueError(err_msg, e) from e

    def __read_language_configuration(self):
        logger.debug("Reading language configuration...")

        lang_conf = self._config["language"] = {}

        lang_conf["locale"] = self._config_parser.get("language", "locale")

        logger.debug("Language configuration correctly read.")

    def __read_twitter_configuration(self):
        logger.debug("Reading Twitter configuration...")

        twitter_conf = self._config["twitter"] = {}

        twitter_conf["bearer_token"] = self._config_parser.get(
            "twitter", "bearer_token"
        )
        twitter_conf["consumer_key"] = self._config_parser.get(
            "twitter", "consumer_key"
        )
        twitter_conf["consumer_secret"] = self._config_parser.get(
            "twitter", "consumer_secret"
        )
        twitter_conf["access_token_key"] = self._config_parser.get(
            "twitter", "access_token_key"
        )
        twitter_conf["access_token_secret"] = self._config_parser.get(
            "twitter", "access_token_secret"
        )

        logger.debug("Twitter configuration correctly read.")

    def __read_postgresql_configuration(self):
        logger.debug("Reading PostgreSQL configuration...")

        postgresql_conf = self._config["postgresql"] = {}

        postgresql_conf["user"] = self._config_parser.get("postgresql", "user")
        postgresql_conf["password"] = self._config_parser.get("postgresql", "password")
        postgresql_conf["hostname"] = self._config_parser.get("postgresql", "hostname")
        postgresql_conf["database"] = self._config_parser.get("postgresql", "database")
        postgresql_conf["ephemeris_table"] = self._config_parser.get(
            "postgresql", "ephemeris_table"
        )
        postgresql_conf["logging_echo"] = self._config_parser.get(
            "postgresql", "logging_echo"
        )

        logger.debug("PostgreSQL configuration correctly read.")

    @property
    def config(self):
        """Returns current config"""
        return self._config



"""Almanac Bot module"""

import argparse
import json
import logging
import logging.config
import os
import sys
from typing import List

from babel import Locale, UnknownLocaleError

from almanacbot import config, constants
from almanacbot.ephemeris import Ephemeris
from almanacbot.postgresql_client import PostgreSQLClient
from almanacbot.twitter_client import TwitterClient

logger = logging.getLogger("almanacbot")


class AlmanacBot:
    """Almanac Bot class"""

    def __init__(self):
        self.conf: config.Configuration = None
        self.twitter_client: TwitterClient = None
        self.postgresql_client: PostgreSQLClient = None

        # configure logger
        self._setup_logging()

        # read configuration
        logger.info("Initializing Almanac Bot...")
        try:
            self.conf = config.Configuration(constants.CONFIG_FILE_NAME)
        except ValueError:
            logger.exception("Error getting configuration.")
            sys.exit(1)

        # setup locale
        try:
            self.locale: Locale = Locale.parse(self.conf.config["language"]["locale"])
            logger.info(f"Locale set to: {self.locale}")
        except (ValueError, UnknownLocaleError):
            logger.exception("Error setting up locale.")
            sys.exit(1)

        # setup Twitter API client
        try:
            self._setup_twitter()
        except ValueError:
            logger.exception("Error setting up Twitter API client.")
            sys.exit(1)

        # setup PostgreSQL client
        try:
            self._setup_postgresql()
        except ValueError:
            logger.exception("Error setting up PostgreSQL client.")
            sys.exit(1)

        logger.info("Almanac Bot properly initialized.")

    def _setup_logging(
        self,
        path="logging.json",
        log_level=logging.DEBUG,
        env_key=constants.CONFIG_ENVVAR,
    ) -> None:
        env_path: str = os.getenv(env_key, None)
        if env_path:
            path = env_path
        if os.path.exists(path):
            with open(path, "rt", encoding="UTF-8") as f:
                log_conf = json.load(f)
                logging.config.dictConfig(log_conf)
                logger.debug(f"Loaded logging configuration:\n{log_conf}")
        else:
            logging.basicConfig(level=log_level)
            logger.debug("Default logging configuration applied.")

    def _setup_twitter(self) -> None:
        logger.info("Setting up Twitter API client...")
        self.twitter_client = TwitterClient(
            bearer_token=self.conf.config["twitter"]["bearer_token"],
            consumer_key=self.conf.config["twitter"]["consumer_key"],
            consumer_secret=self.conf.config["twitter"]["consumer_secret"],
            access_token_key=self.conf.config["twitter"]["access_token_key"],
            access_token_secret=self.conf.config["twitter"]["access_token_secret"],
            locale=self.locale,
        )
        logger.info("Twitter API client set up.")

    def _setup_postgresql(self) -> None:
        logger.info("Setting up PostgreSQL client...")
        self.postgresql_client: PostgreSQLClient = PostgreSQLClient(
            user=self.conf.config["postgresql"]["user"],
            password=self.conf.config["postgresql"]["password"],
            hostname=self.conf.config["postgresql"]["hostname"],
            database=self.conf.config["postgresql"]["database"],
            ephemeris_table=self.conf.config["postgresql"]["ephemeris_table"],
            logging_echo=bool(self.conf.config["postgresql"]["logging_echo"]),
        )
        logger.info("PostgreSQL client set up.")

    def run(self, dry_run: bool = False) -> int:
        """
        Execute once: get untweeted ephemeris for today, tweet them.

        Args:
            dry_run: If True, log what would be tweeted without actually tweeting.

        Returns:
            Number of tweets sent (or would be sent in dry-run mode).
        """
        logger.info("Getting today's untweeted ephemeris...")
        today_ephs: List[Ephemeris] = (
            self.postgresql_client.get_untweeted_today_ephemeris()
        )

        if not today_ephs:
            logger.info("No untweeted ephemeris for today.")
            return 0

        logger.debug(f"Found {len(today_ephs)} untweeted ephemeris entries.")

        tweets_sent = 0
        for eph in today_ephs:
            try:
                if dry_run:
                    text = TwitterClient._process_tweet_text(eph, self.locale)
                    logger.info(f"[DRY-RUN] Would tweet id={eph.id}: {text}")
                else:
                    logger.info(f"Tweeting ephemeris id={eph.id}...")
                    self.twitter_client.tweet_ephemeris(eph)
                    self.postgresql_client.mark_as_tweeted(eph.id)
                    logger.info(f"Successfully tweeted ephemeris id={eph.id}")
                tweets_sent += 1
            except Exception:
                logger.exception(f"Failed to tweet ephemeris id={eph.id}")

        mode = "would be sent" if dry_run else "sent"
        logger.info(f"Completed: {tweets_sent}/{len(today_ephs)} tweets {mode}.")
        return tweets_sent


def main() -> None:
    """Main entry point for one-shot execution."""
    parser = argparse.ArgumentParser(
        description="Almanac Bot - Tweet historical events"
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Log tweets without actually sending them",
    )
    args = parser.parse_args()

    logger.info("Starting Almanac Bot (one-shot mode)...")

    ab = AlmanacBot()
    tweets_sent = ab.run(dry_run=args.dry_run)

    mode = "would be" if args.dry_run else ""
    logger.info(f"Almanac Bot finished. Tweets {mode} sent: {tweets_sent}")
    sys.exit(0)


if __name__ == "__main__":
    main()



"""Init file of the module"""

__all__ = [
    "almanacbot",
    "config",
    "constants",
    "ephemeris",
    "twitter_client",
    "postgresql_client",
    "data_loader",
]

from almanacbot import (
    almanacbot,
    config,
    constants,
    ephemeris,
    twitter_client,
    postgresql_client,
    data_loader,
)



import datetime
from dataclasses import dataclass
from typing import Optional

from sqlalchemy import TIMESTAMP, Text
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column
from sqlalchemy.types import UserDefinedType
import sqlalchemy


class Base(DeclarativeBase):
    pass


@dataclass
class Location:
    latitude: float
    longitude: float


class LatLngType(UserDefinedType):
    """
    Custom SQLAlchemy type to handle POINT columns.

    References:

    - https://gist.github.com/kwatch/02b1a5a8899b67df2623
    - https://docs.sqlalchemy.org/en/14/core/custom_types.html#sqlalchemy.types.UserDefinedType  # noqa
    """

    # Can do because we made the Location dataclass hashable
    cache_ok = True

    def get_col_spec(self):
        return "POINT"

    def bind_expression(self, bindvalue):
        return sqlalchemy.func.POINT(bindvalue, type_=self)

    def bind_processor(self, dialect):
        """
        Return function to serialize a Location into a database string literal.
        """

        def process(value: Location | tuple[float, float] | None) -> str | None:
            if value is None:
                return None

            if isinstance(value, tuple):
                value = Location(*value)

            return f"({value.latitude},{value.longitude})"

        return process

    def result_processor(self, dialect, coltype):
        """
        Return function to parse a database string result into Python data type.
        """

        def process(value: str) -> Location | None:
            if value is None:
                return None

            latitude, longitude = value.strip("()").split(",")

            return Location(float(latitude), float(longitude))

        return process


@dataclass
class Ephemeris(Base):
    __tablename__ = "ephemeris"

    id: Mapped[int] = mapped_column(primary_key=True)
    date: Mapped[datetime.datetime] = mapped_column(TIMESTAMP(timezone=True))
    text: Mapped[str] = mapped_column(Text)
    location: Mapped[Optional[Location]] = mapped_column(LatLngType, default=None)
    last_tweeted_at: Mapped[Optional[datetime.datetime]] = mapped_column(
        TIMESTAMP(timezone=True), default=None
    )



import datetime
from typing import List

import sqlalchemy
from sqlalchemy import (
    Select,
    and_,
    create_engine,
    extract,
    func,
    insert,
    null,
    or_,
    select,
)
from sqlalchemy.orm import Session

from almanacbot.ephemeris import Ephemeris


class PostgreSQLClient:
    """Class serving as PostgreSQL client"""

    def __init__(
        self,
        user: str,
        password: str,
        hostname: str,
        database: str,
        ephemeris_table: str,
        logging_echo: bool,
    ):
        self.engine = create_engine(
            f"postgresql+psycopg://{user}:{password}@{hostname}/{database}",
            echo=logging_echo,
        )
        self.ephemeris_table: str = ephemeris_table

    def get_today_ephemeris(self) -> List[Ephemeris]:
        """Get all ephemeris entries for today using month+day matching."""
        now = sqlalchemy.func.now()
        query: Select = select(Ephemeris).filter(
            and_(
                extract("MONTH", Ephemeris.date) == extract("MONTH", now),
                extract("DAY", Ephemeris.date) == extract("DAY", now),
            )
        )
        with Session(self.engine) as session:
            ephs: List[Ephemeris] = session.scalars(query).all()
            return ephs

    def get_untweeted_today_ephemeris(self) -> List[Ephemeris]:
        """
        Get ephemeris entries for today that haven't been tweeted yet today.

        Uses month+day matching (handles leap years correctly) and checks
        that last_tweeted_at is either NULL or before today (idempotency).
        """
        now = func.now()
        today_start = func.date_trunc("day", now)

        query: Select = select(Ephemeris).filter(
            and_(
                extract("MONTH", Ephemeris.date) == extract("MONTH", now),
                extract("DAY", Ephemeris.date) == extract("DAY", now),
                or_(
                    Ephemeris.last_tweeted_at.is_(None),
                    Ephemeris.last_tweeted_at < today_start,
                ),
            )
        )
        with Session(self.engine) as session:
            ephs: List[Ephemeris] = session.scalars(query).all()
            return ephs

    def mark_as_tweeted(self, ephemeris_id: int) -> None:
        """Mark an ephemeris entry as tweeted with current UTC timestamp."""
        with Session(self.engine) as session:
            eph = session.get(Ephemeris, ephemeris_id)
            if eph:
                eph.last_tweeted_at = datetime.datetime.now(datetime.timezone.utc)
                session.commit()

    def count_ephemeris(self) -> int:
        with Session(self.engine) as session:
            stmnt = func.count(Ephemeris.id)
            return session.execute(stmnt).scalar()

    def insert_ephemeris(self, eph: Ephemeris):
        with Session(self.engine) as session:
            stmnt = insert(Ephemeris).values(
                date=eph.date,
                text=eph.text,
                location=(
                    sqlalchemy.func.point(eph.location.latitude, eph.location.longitude)
                    if eph.location is not None
                    else null()
                ),
            )
            session.execute(stmnt)
            session.commit()



import datetime
import logging
import string
# from typing import List

from babel import Locale
from babel.dates import format_date
import tweepy

from almanacbot.ephemeris import Ephemeris

logger = logging.getLogger(__name__)


class TwitterClient:
    """Class serving as Twitter API client"""

    def __init__(
        self,
        bearer_token: str,
        consumer_key: str,
        consumer_secret: str,
        access_token_key: str,
        access_token_secret: str,
        locale: Locale,
    ):
        self.locale = locale

        # Twitter API v2 client
        self._client_v2: tweepy.Client = tweepy.Client(
            bearer_token=bearer_token,
            consumer_key=consumer_key,
            consumer_secret=consumer_secret,
            access_token=access_token_key,
            access_token_secret=access_token_secret,
        )

        # Twitter API v1 client
        # self._client_v1: tweepy.API = tweepy.API(tweepy.OAuth2BearerHandler(bearer_token))

    def tweet_ephemeris(self, eph: Ephemeris) -> None:
        # tplace: tweepy.Place = None
        # no access to places using Twitter v1 API with free account
        # if eph.location:
        #     logger.info(f"Obtaining Place from ephemeris coordinates: {eph.location}")
        #     places: List[tweepy.Place] = self._client_v1.search_geo(
        #         lat=eph.location.latitude,
        #         lon=eph.location.longitude,
        #         max_results=1,
        #     )
        #     if len(places) > 0:
        #         tplace = places[0]

        # logger.info(f"Tweeting ephemeris: {eph}")
        # self._client_v2.create_tweet(
        #     text=TwitterClient._process_tweet_text(eph, self.locale),
        #     place_id=tplace[0] if tplace else None,
        # )

        # post tweet without place ID (geolocation)
        logger.info(f"Tweeting ephemeris: {eph}")
        self._client_v2.create_tweet(
            text=TwitterClient._process_tweet_text(eph, self.locale),
        )

    @staticmethod
    def _process_tweet_text(eph: Ephemeris, locale: Locale) -> str:
        today = datetime.datetime.now(datetime.timezone.utc)
        template = string.Template(eph.text)

        values = {
            "date": format_date(date=eph.date.date(), format="full", locale=locale),
            "years_ago": today.year - eph.date.year,
        }

        logger.debug(f"Processed ephemeris text: {template.substitute(values)}")

        return template.substitute(values)



import ast
import configparser
import csv

from psycopg import OperationalError
import typer

from almanacbot import constants
from almanacbot.config import Configuration
from almanacbot.ephemeris import Ephemeris, Location
from almanacbot.postgresql_client import PostgreSQLClient

config_parser: configparser = configparser.ConfigParser()


def read_configuration():
    # read configuration
    print("Reading configuration...")
    try:
        return Configuration(constants.CONFIG_FILE_NAME).config
    except ValueError:
        print("Error getting configuration: {err}")
        typer.Exit(1)
    print("Configuration correctly read.")


def main(
    csv_file_path: str = "init_db.csv",
):
    config: dict = read_configuration()

    with open(csv_file_path, "r") as csv_file:
        try:
            print("Connecting to PostgreSQL...")
            psql_client = PostgreSQLClient(
                user=config["postgresql"]["user"],
                password=config["postgresql"]["password"],
                hostname=config["postgresql"]["hostname"],
                database=config["postgresql"]["database"],
                ephemeris_table=config["postgresql"]["ephemeris_table"],
                logging_echo=bool(config["postgresql"]["logging_echo"]),
            )

            print("Checking for existing data...")
            if psql_client.count_ephemeris() > 0:
                confirmation: str = typer.confirm(
                    "There is data in the databse, are you sure you want to append more?"
                )
                if not confirmation:
                    print("Aborting!")
                    raise typer.Abort()

            print("Reading and inserting data...")
            csvReader = csv.reader(csv_file, delimiter=";")
            next(csvReader, None)  # skip header
            for row in csvReader:
                print(f"Read row: {row}")
                location_tpl: tuple = ast.literal_eval(row[2]) if row[2] else None
                location: Location = None
                if location_tpl:
                    location: Location = Location(
                        latitude=location_tpl[0],
                        longitude=location_tpl[1],
                    )
                psql_client.insert_ephemeris(
                    Ephemeris(
                        date=row[0],
                        text=row[1],
                        location=location,
                    )
                )
        except (OperationalError, ValueError) as exc:
            print(f"Error introducing CSV data to the DB: {exc}")
            typer.Exit(2)
